;*********************************************************************************************************
;**********	Rechenoperationen zum Tapetimer Version 2.1					**********
;*********************************************************************************************************

;*********************************************************************************************************
;**********	Umwandlung 8 Bit Binrzahl zu 2-stelligem gepacktem BCD. Eingangswert var10	**********
;**********	wird zerstrt. Ergebnis in var10; eingeschrnkter Wertebereich 0 bis 99		**********
;**********	Number of cycles: 5/50 (Min/Max) + return					**********
;**********	benutzte Register: var1[1:0]							**********
;*********************************************************************************************************
bin2bcd8:
	clr	var11			;clear result
bBCD8_1:
	subi	var10,10		;input = input - 10
	brcs	bBCD8_2			;abort if carry set
	subi	var11,-$10 		;tBCDH = tBCDH + 10
	rjmp	bBCD8_1			;loop again
bBCD8_2:
	subi	var10,-10		;compensate extra subtraction
	add	var10,var11		;
	ret

;*********************************************************************************************************
;**********	Umwandlung 16 Bit Binrzahl var1[1:0] in 5-stelliges gepacktes BCD		**********
;**********											**********
;**********	Title:	32-bit Arithmetic Routines with Macrolibrary				**********
;**********	Originator:	(C) 1999-2003 Andre Birua (mailto:birua@hotmail.com)		**********
;**********			This Application Note absolutely free in use anybody		**********
;**********	Number of cycles 22/180 (Min/Max)						**********
;**********	Equivalent of AVR204, but smaller & much faster					**********
;**********	benutzte Register: 	var1[1:0], var2[2:0], tBCD[2:0]				**********
;**********				tBCD2 = BCD Wert digit  1				**********
;**********				tBCD1 = BCD Wert digits 2 und 3	  			**********
;**********				tBCD0 = BCD Wert digits 4 und 5				**********
;*********************************************************************************************************
bin2bcd16:
	ldi     var22,0xff		;initialize digit 4
binbcd_4:
	inc     var22			;
	subi    var10,byte1(10000)	;subiw fbin,10000
	sbci    var11,byte2(10000)	;
	brcc    binbcd_4		;
	ldi     var21,0x9f		;initialize digits 3 and 2
binbcd_3:
	subi    var21,0x10		;
	subi    var10,byte1(-1000)	;addiw fbin,1000
	sbci    var11,byte2(-1000)	;
	brcs    binbcd_3		;
binbcd_2:
	inc     var21			;
	subi    var10,byte1(100)	;subiw fbin,100
	sbci    var11,byte2(100)	;
	brcc    binbcd_2		;
	ldi     var20,0xa0		;initialize digits 1 and 0
binbcd_1:
	subi    var20,0x10		;
	subi    var10,-10		;addi fbin,10
	brcs    binbcd_1		;
	add     var20,var10		;LSD
binbcd_ret:
	mov	tBCD0,var20		;Ergebnis sichern
	mov	tBCD1,var21		;um die High-Register wieder frei zu bekommen
	mov	tBCD2,var22		;
	ret

;*********************************************************************************************************
;**********	Umwandlung 16-Bit Sekundenlaufzeit var10[1:0] ins 4-stellige Format MM:SS	**********
;**********	oder ins 5-stellige Format H:MM:SS als gepacktes BCD				**********
;**********	benutzte Register: var1[2:0], var2[1:0]						**********
;**********	tBCD2 = BCD Wert H					  			**********
;**********	tBCD1 = BCD Wert MM					  			**********
;**********	tBCD0 = BCD Wert SS					  			**********
;*********************************************************************************************************
sek2min2BCD:
	clr	var20			;lsche Ergebnisregister Minuten und
	clr	var21			;Stunden
teste_minuten:
	cpi	var10,byte1(60)		;Sekundenlaufzeit grer 60?
	ldi	var12,byte2(60)		;
	cpc	var11,var12		;
	brlo	wandle_sekunden		;nein keine Minute mehr drauf
	subi	var10,byte1(60)		;sonst 60 von Sekundenlaufzeit
	sbci	var11,byte2(60)		;abziehen und
	inc	var20			;eins auf Minutenlaufzeit draufaddieren
.if (anzeigestellen==4)			;nur bei 4 Digits
	cpi	var20,100		;Minutenlaufzeit jetzt grer 100? 
.else					;nur bei 5 Digits
	cpi	var20,60		;Minutenlaufzeit ber 60?
.endif
	brlo	teste_minuten		;nein, alles noch erlaubt
.if (anzeigestellen==4)			;nur bei 4 Digits
	subi	var20,100		;sonst 100 abziehen
.else					;nur bei 5 Digits
	subi	var20,60		;sonst 60 von Minutenlaufzeit abziehen
.endif
	inc	var21			;eins auf die Stunden draufrechnen
	rjmp	teste_minuten		;auf weitere Minute testen
wandle_sekunden:			;in [var21:var20:var10] stehen nun Std:Min:Sek
	rcall	bin2bcd8		;nach BCD umformen
	mov	tBCD0,var10		;in tBCD0 sichern
	mov	var10,var20		;Minuten holen
	rcall	bin2bcd8		;nach BCD umformen
	mov	tBCD1,var10		;in tBCD1 sichern
	mov	var10,var21		;Stunden holen
	rcall	bin2bcd8		;nach BCD umformen
	mov	tBCD2,var10		;in tBCD2 sichern
	ret

;*********************************************************************************************************
;**********	16Bit Hardware-Multiplikation							**********
;**********	bentigte Register: Eingangswerte = var1[1:0] und var2[1:0]			**********
;**********				 Ergebnis = mod[3:0]					**********
;**********	Hilfsregister = r0, r1, Eingangsregister werden nicht zerstrt			**********
;**********	http://www.avr-asm-tutorial.net/avr_de/rechnen/hardmult.html			**********
;*********************************************************************************************************
mul16u:
	mul	var11,var21 		;Multipliziere MSBs
	mov	mod2,r0 		;Kopie in obere Ergebnisregister
	mov	mod3,r1			;
	mul	var10,var20 		;Multipliziere LSBs
	mov	mod0,r0 		;Kopie in untere Ergebnisregister
	mov	mod1,r1			;
	mul	var11,var20 		;Multipliziere 1M mit 2L
	add	mod1,r0 		;Addiere zum Ergebnis
	adc	mod2,r1			;
	adc	mod3,null 		;Addiere Uebertrag
	mul	var10,var21 		;Multipliziere 1L mit 2M
	add	mod1,r0 		;Addiere zum Ergebnis
	adc	mod2,r1			;
	adc	mod3,null		;
	ret

;*********************************************************************************************************
;**********	16Bit * 24Bit Hardware-Multiplikation						**********
;**********	bentigte Register: Eingangswerte = var1[1:0] und var2[2:0]			**********
;**********				 Ergebnis = mod[4:0]					**********
;**********	Hilfsregister = r0, r1, Eingangsregister werden nicht zerstrt			**********
;**********	http://www.avr-asm-tutorial.net/avr_de/rechnen/hardmult.html			**********
;*********************************************************************************************************
mul16x24u:
	clr	mod0			;Ergebnisregister lschen
	clr	mod1			;
	clr	mod2			;
	clr	mod3			;
	clr	mod4			;

	mul	var11,var22		;Term 1
	add	mod3,r0			;addiere zum Ergebnis
	adc	mod4,r1			;
	mul	var11,var21		;Term 2
	add	mod2,r0			;
	adc	mod3,r1			;
	adc	mod4,null		;(addiere moeglichen Ueberlauf)
	mul	var11,var20		;Term 3
	add	mod1,r0			;
	adc	mod2,r1			;
	adc	mod3,null		;
	adc	mod4,null		;
	mul	var10,var22		;Term 4
	add	mod2,r0			;
	adc	mod3,r1			;
	adc	mod4,null		;
	mul	var10,var21		;Term 5
	add	mod1,r0			;
	adc	mod2,r1			;
	adc	mod3,null		;
	adc	mod4,null		;
	mul	var10,var20		;Term 6
	add	mod0,r0			;
	adc	mod1,r1			;
	adc	mod2,null		;
	adc	mod3,null		;
	adc	mod4,null		;
	ret

;*********************************************************************************************************
;**********	32Bit vorzeichenlose Division	von http://elm-chan.org/cc_e.html		**********
;**********	bentigte Register: var1[3:0] = dividend (0x000000..0xffffff)			**********
;**********	                    var2[3:0] = divisor  (0x000001..0x7fffff)			**********
;**********			     mod[3:0], lc						**********
;**********	Ergebnis:           var1[3:0]		Clock = 549...677 cycles (+ret)		**********
;*********************************************************************************************************
div32u:
	clr	mod0			;initialize variables
	clr	mod1			;  mod = 0;
	clr	mod2			;  lc = 32;
	clr	mod3			;
	ldi	lc,32			;/
;---- calculating loop
	lsl	var10			;var1 = var1 << 1;
	rol	var11			;
	rol	var12			;
	rol	var13			;/
	rol	mod0			;mod = mod << 1 + carry;
	rol	mod1			;
	rol	mod2			;
	rol	mod3			;/
	cp	mod0,var20		;if (mod => var2) {
	cpc	mod1,var21		; mod -= var2; var1++;
	cpc	mod2,var22		; }
	cpc	mod3,var23		;
	brcs	PC+6			;
	inc	var10			;
	sub	mod0,var20		;
	sbc	mod1,var21		;
	sbc	mod2,var22		;
	sbc	mod3,var23		;/
	dec	lc			;if (--lc > 0)
	brne	PC-19			; continue loop;
	ret

;*********************************************************************************************************
;**********	Fast AVR integer square root assembly routine By Ruud v Gessel			**********
;**********	var1[1:0] = SQRT(mod[3:0]) rounded down to integer				**********
;**********	Cycles incl call & ret = 259 - 301						**********
;**********	benutzte Register: var1[3:0], destroys the argument in mod[3:0]			**********
;*********************************************************************************************************
Sqrt32:
	ldi	var13,0xc0
	clr	var12			;rotation mask in R19:R18
	ldi	var11,0x40
	sub	var10,var10		;developing sqrt in R17:R16, C=0 
_sq32_1:
	brcs 	_sq32_2			;C --> Bit is always 1
	cp	mod2,var10
	cpc	mod3,var11		;Does test value fit?
	brcs	_sq32_3			;C --> nope, bit is 0
_sq32_2:
	sub	mod2,var10
	sbc	mod3,var11		;Adjust argument for next bit
	or	var10,var12
	or	var11,var13		;Set bit to 1
_sq32_3:
	lsr	var13
	ror	var12			;Shift right mask, C --> end loop
	eor	var11,var13
	eor	var10,var12		;Shift right only test bit in result
	rol	mod0			;Bit 0 only set if end of loop
	rol	mod1
	rol	mod2
	rol	mod3			;Shift left remaining argument (C used at _sq32_1)
	sbrs	mod0,0			;Skip if 15 bits developed
	rjmp	_sq32_1			;Develop 15 bits of the sqrt 
 
	brcs	_sq32_4			;C--> Last bits always 1
	lsl	mod1			;Need bit 7 in C for cpc
	cpc	var10,mod2
	cpc	var11,mod3		;After this C is last bit
_sq32_4:
	adc	var10,var13		;Round up if C (R19=0) 
	ret
